####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
The :obj:`.utils` module provides useful utility functions.
"""
import os
import sys
import natsort
import datetime
import numpy as np


class Logger:
    """
    Redirects the STDOUT and STDERR to a specified output file while preserving them on screen.

    Parameters
    ----------
    logfile : str
        Name of the output file to write the logged messages.

    Attributes
    ----------
    terminal : file object
        The file object that represents the original STDOUT (i.e., the screen).
    log : file object
        The file object that represents the logfile where messages will be written.
    """

    def __init__(self, logfile):
        """
        Initializes a Logger instance.

        Parameters
        ----------
        logfile : str
            Name of the output file to write the logged messages.
        """
        self.terminal = sys.stdout
        self.log = open(logfile, "a")

    def write(self, message):
        """
        Writes the given message to both the STDOUT and the logfile.

        Parameters
        ----------
        message : str
            The message to be written to STDOUT and logfile.
        """
        self.terminal.write(message)
        self.log.write(message)

    def flush(self):
        """
        This method is needed for Python 3 compatibility. This handles the flush command by doing nothing.
        You might want to specify some extra behavior here.
        """
        # self.terminal.log()
        pass


def gmx_output(gmx_obj):
    """
    Tells if a command launched by ``gmxapi.commandline_operation`` failed or not.
    If the command failed, the function will print out STDOUT and STDERR of the process.

    Parameters
    ----------
    gmx_obj : obj
        An object generated by gmx.commandline_operation.
    """
    try:
        rtn_code = gmx_obj.output.returncode.result()
        if type(rtn_code) == list:  # simulation ensemble
            if sum(rtn_code) == 0:
                print(" DONE")
            else:
                print(f" Return codes: {rtn_code}")
                for i in range(len(rtn_code)):
                    if rtn_code[i] != 0:
                        print(f"\nSTDERR of the process:\n\n {gmx_obj.output.stderr.result()[i]}\n")
        else:  # single simulation
            if rtn_code == 0:
                print(" DONE")
            else:
                print(f"\nSTDERR of the process:\n\n {gmx_obj.output.stderr.result()}\n")
    except AttributeError:
        raise RuntimeError("The input object is not an object generated by commandline_operation.")


def clean_up(dir_before, dir_after, verbose=False):
    """
    Removes newly generated folders that are empty after a command launched by gmxapi.

    Parameters
    ----------
    dir_before : list
        The list of directories existed before the command was executed.
    dir_after : list
        The list of directories existed after the command was executed. This helps
        figure out which directories are newly generated by command.
    verbose : bool
        If details should be printed.
    """
    new_dirs = natsort.natsorted([i for i in dir_after if i not in dir_before])
    if len(new_dirs) != 0:
        if verbose is True:
            print("\nCleaning up/restructuring the directories ...")
    for i in new_dirs:
        if len(os.listdir(i)) == 0:
            if verbose is True:
                print(f"  Removing the empty folder {i} ...")
            os.rmdir(i)


def format_time(t):
    """
    Converts time in seconds to the "most readable" format.

    Parameters
    ----------
    t : float
        The time in seconds.

    Returns
    -------
    t_str : sttr
        A string in the format of "XX day XX hour(s) XX minute(s) XX second(s)".
    """
    hh_mm_ss = str(datetime.timedelta(seconds=t)).split(":")

    if "day" in hh_mm_ss[0]:
        # hh_mm_ss[0] will contain "day" and cannot be converted to float
        hh, mm, ss = hh_mm_ss[0], float(hh_mm_ss[1]), float(hh_mm_ss[2])
        t_str = f"{hh} hour(s) {mm:.0f} minute(s) {ss:.0f} second(s)"
    else:
        hh, mm, ss = float(hh_mm_ss[0]), float(hh_mm_ss[1]), float(hh_mm_ss[2])
        if hh == 0:
            if mm == 0:
                t_str = f"{ss:.1f} second(s)"
            else:
                t_str = f"{mm:.0f} minute(s) {ss:.0f} second(s)"
        else:
            t_str = f"{hh:.0f} hour(s) {mm:.0f} minute(s) {ss:.0f} second(s)"

    return t_str


def autoconvert(s):
    """
    Converts input to a numerical type if possible. Used for the MDP parser.
    Modified from `utilities.py in GromacsWrapper <https://github.com/Becksteinlab/GromacsWrapper>`_.
    Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>

    Parameters
    ----------
    s : str or any
        The input value to be converted to a numerical type if possible. If :code:`s` is not a string,
        it is returned as is.

    Returns
    -------
    numerical : int, float, numpy.ndarray, or any
        The converted numerical value. If :code:`s` can be converted to a single numerical value,
        that value is returned as an :code:`int` or :code:`float`. If :code:`s` can be converted to
        multiple numerical values, a :code:`numpy.ndarray` containing those values is returned.
        If :code:`s` cannot be converted to a numerical value, :code:`s` is returned as is.

    Raises
    ------
    ValueError
        If :code:`s` cannot be converted to a numerical value.
    """
    if type(s) is not str:
        return s
    for converter in int, float, str:  # try them in increasing order of lenience
        try:
            s = [converter(i) for i in s.split()]
            if len(s) == 1:
                return s[0]
            else:
                return np.array(s)
        except (ValueError, AttributeError):
            pass
    raise ValueError("Failed to autoconvert {0!r}".format(s))


def get_subplot_dimension(n_panels):
    """
    Gets the numbers of rows and columns in a subplot such that
    the arrangement of the .

    Parameters
    ----------
    n_panels : int
        The number of panels in the subplot.
    """
    if int(np.sqrt(n_panels) + 0.5) ** 2 == n_panels:
        # perfect square number
        n_cols = int(np.sqrt(n_panels))
    else:
        n_cols = int(np.floor(np.sqrt(n_panels))) + 1

    if n_panels % n_cols == 0:
        n_rows = int(np.floor(n_panels / n_cols))
    else:
        n_rows = int(np.floor(n_panels / n_cols)) + 1

    return n_rows, n_cols


def weighted_mean(vals, errs):
    """
    Calculates the inverse-variance-weighted mean.

    Parameters
    ----------
    vals : list
        A list of values to be averaged.
    errs : list
        A list of errors corresponding to the given values

    Returns
    -------
    mean : float
        The inverse-variance-weighted mean.
    err : float
        The propgated error of the mean.
    """
    w = [1 / (i ** 2) for i in errs]
    wx = [w[i] * vals[i] for i in range(len(vals))]
    mean = np.sum(wx) / np.sum(w)
    err = np.sqrt(1 / np.sum(w))

    return mean, err



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
The :obj:`.analyze_free_energy` module provides functions for performing free energy calculations for EEXE simulations.
"""
import alchemlyb
import numpy as np
import matplotlib.pyplot as plt
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

from pymbar.timeseries import detect_equilibration, subsample_correlated_data  # noqa: E402
from alchemlyb.estimators import TI, BAR, MBAR  # noqa: E402
from alchemlyb.parsing.gmx import extract_dHdl, extract_u_nk  # noqa: E402
from alchemlyb.preprocessing import subsampling  # noqa: E402
from ensemble_md.utils import utils  # noqa: E402
from ensemble_md.utils.exceptions import ParameterError  # noqa: E402


def preprocess_data(files, temp, spacing=1, get_u_nk=True, get_dHdl=False):
    """
    This function preprocesses :math:`dH/dλ` data obtained from the EEXE simulation.
    For each replica, it reads in :math:`dH/dλ` data from all iterations, concatenate
    them, remove the equilibrium region and and decorrelate the concatenated data. Notably,
    the data preprocessing protocol is basically the same as the one adopted in
    :code:`alchemlyb.subsampling.equilibrium_detection`.

    Parameters
    ----------
    files : list
        A list of naturally sorted dhdl files from all iterations of one replica.
    temp : float
        The simulation temperature in Kelvin.
    spacing : int
        The spacing (number of data points) to consider when subsampling the data.
    get_u_nk : bool
        Whether to get the u_nk data from the dhdl files. The default is True.
    get_dHdl : bool
        Whether to get the dHdl data from the dhdl files. the default is False.

    Returns
    -------
    preprocessed_u_nk : pd.Dataframe
        The preprocessed dHdl data that can serve as the input to free energy estimators.
    preprocessed_dHdl : pd.Dataframe
        The preprocessed dHdl data that can serve as the input to free energy estimators.
    """
    if get_u_nk is True:
        print('Collecting u_nk data from all iterations ...')
        u_nk = alchemlyb.concat([extract_u_nk(xvg, T=temp) for xvg in files])
        u_nk_series = subsampling.u_nk2series(u_nk)  # default method: 'dE'
        u_nk, u_nk_series = subsampling._prepare_input(u_nk, u_nk_series, drop_duplicates=True, sort=True)
        u_nk = subsampling.slicing(u_nk, step=spacing)
        u_nk_series = subsampling.slicing(u_nk_series, step=spacing)

        print('Subsampling and decorrelating the concatenated u_nk data ...')
        t, statinef, Neff_max = detect_equilibration(u_nk_series.values)

        print(f'  Adopted spacing: {spacing: .0f}')
        print(f' {t / len(u_nk_series) * 100: .1f}% of the u_nk data was in the equilibrium region and therfore discarded.')  # noqa: E501
        print(f'  Statistical inefficiency of u_nk: {statinef: .1f}')
        print(f'  Number of effective samples: {Neff_max: .0f}\n')

        u_nk_series_equil, u_nk_equil = u_nk_series[t:], u_nk[t:]
        indices = subsample_correlated_data(u_nk_series_equil, g=statinef)
        preprocessed_u_nk = u_nk_equil.iloc[indices]

    else:
        preprocessed_u_nk = None

    if get_dHdl is True:
        print('Collecting dHdl data from all iterations ...')
        dHdl = alchemlyb.concat([extract_dHdl(xvg, T=temp) for xvg in files])
        dHdl_series = subsampling.dhdl2series(dHdl)  # default method: 'dE'
        dHdl, dHdl_series = subsampling._prepare_input(dHdl, dHdl_series, drop_duplicates=True, sort=True)
        dHdl = subsampling.slicing(dHdl, step=spacing)
        dHdl_series = subsampling.slicing(dHdl_series, step=spacing)

        print('Subsampling and decorrelating the concatenated dHdl data ...')
        t, statinef, Neff_max = detect_equilibration(dHdl_series.values)

        print(f'  Adopted spacing: {spacing: .0f}')
        print(f' {t / len(dHdl_series) * 100: .1f}% of the dHdl data was in the equilibrium region and therfore discarded.')  # noqa: E501
        print(f'  Statistical inefficiency of dHdl: {statinef: .1f}')
        print(f'  Number of effective samples: {Neff_max: .0f}\n')

        dHdl_series_equil, dHdl_equil = dHdl_series[t:], dHdl[t:]
        indices = subsample_correlated_data(dHdl_series_equil, g=statinef)
        preprocessed_dHdl = dHdl_equil.iloc[indices]

    else:
        preprocessed_dHdl = None

    return preprocessed_u_nk, preprocessed_dHdl


def _calculate_df_adjacent(data, df_method="MBAR"):
    """
    An Internal function that generates a list of estimators fitting the input data
    and calculates at list of free energy between adjacent states for all replicas.

    Parameters
    ----------
    data : pd.Dataframe
        A list of dHdl or u_nk dataframes obtained from all replicas of the EEXE simulation of interest.
        Preferrably, the dHdl or u_nk data should be preprocessed by the function proprocess_data.

    Returns
    -------
    df_adjacent : list
        A list of free energy differences between adjacent states for all replicas.
    df_err_adjacent : list
        A list of uncertainties corresponding to the values of df_adjacent.
    """
    n_sim = len(data)
    estimators = []  # A list of objects of the corresponding class in alchemlyb.estimators
    for i in range(n_sim):
        if df_method == "TI":
            estimators.append(TI().fit(data[i]))
        elif df_method == "BAR":
            estimators.append(BAR().fit(data[i]))
        elif df_method == "MBAR":
            estimators.append(MBAR().fit(data[i]))
        else:
            raise ParameterError('Specified estimator not available.')

    df_adjacent = [list(np.array(estimators[i].delta_f_)[:-1, 1:].diagonal()) for i in range(n_sim)]
    df_err_adjacent = [list(np.array(estimators[i].d_delta_f_)[:-1, 1:].diagonal()) for i in range(n_sim)]

    return df_adjacent, df_err_adjacent


def _calculate_weighted_df(df_adjacent, df_err_adjacent, state_ranges, propagated_err=True):
    """
    An internal function that calculates a list of free energy differences between states i and i + 1.
    For free energy differences obtained from multiple replicas, an average weighted over all involved
    replicas is reported.

    Parameters
    ----------
    df_adjacent : list
        A list of free energy differences between adjacent states for all replicas.
    df_err_adjacent : list
        A list of uncertainties corresponding to the values of df_adjacent.
    state_ranges : list
        A list of lists of intergers that represents the alchemical states that can be sampled by different replicas.
    propagated_err : bool
        Whether to calculate the propagated error when taking the weighted averages for the free energy
        differences that can be obtained from multiple replicas. If False is specified, :code:`df_err`
        returned will be :code:`None`.

    Returns
    -------
    df : list
        A list of free energy differences between states i and i + 1.
    df_err : list
        A list of uncertainties of the free energy differences.
    overlap_bool : list
        overlap_bool[i] = True means that the i-th free energy difference (i.e. df[i]) was available
        in multiple replicas.
    """
    n_tot = state_ranges[-1][-1] + 1
    df, df_err, overlap_bool = [], [], []
    for i in range(n_tot - 1):
        # df_list is a list of free energy difference between sates i and i+1 in different replicas
        # df_err_list contains the uncertainties corresponding to the values of df_list
        df_list, df_err_list = [], []
        for j in range(len(state_ranges)):   # len(state_ranges) = n_sim
            if i in state_ranges[j] and i + 1 in state_ranges[j]:
                idx = state_ranges[j].index(i)
                df_list.append(df_adjacent[j][idx])
                df_err_list.append(df_err_adjacent[j][idx])
        overlap_bool.append(len(df_list) > 1)

        mean, error = utils.weighted_mean(df_list, df_err_list)
        df.append(mean)
        df_err.append(error)

    if propagated_err is False:
        df_err = None

    return df, df_err, overlap_bool


def calculate_free_energy(data, state_ranges, df_method="MBAR", err_method='propagate', n_bootstrap=None, seed=None):
    """
    Caculates the averaged free energy profile with the chosen method given dHdl or u_nk data obtained from
    all replicas of the EEXE simulation of interest. Available methods include TI, BAR, and MBAR. TI
    requires dHdl data while the other two require u_nk data.

    Parameters
    ----------
    data : pd.Dataframe
        A list of dHdl or u_nk dataframes obtained from all replicas of the EEXE simulation of interest.
        Preferrably, the dHdl or u_nk data should be preprocessed by the function proprocess_data.
    state_ranges : list
        A list of lists of intergers that represents the alchemical states that can be sampled by different replicas.
    df_method : str
        The method used to calculate the free energy profile. Available choices include "TI", "BAR", and "MBAR".
    err_method : str
        The method used to estimate the uncertainty of the free energy combined across multiple replicas.
        Available options include "propagate" and "bootstrap". The bootstrapping method is more accurate
        but much more computationally expensive than simple error propagation.
    n_bootstrap : int
        The number of bootstrap iterations. This parameter is used only when the boostrapping method is chosen to
        estimate the uncertainties of the free energies.
    seed : int
        The random seed for bootstrapping.

    Returns
    -------
    f : list
        The full-range free energy profile.
    f_err : list
        The uncertainty corresponding to the values in :code:`f`.
    estimators : list
        A list of estimators fitting the input data for all replicas. With this, the user
        can access all the free energies and their associated uncertainties for all states and replicas.
    """
    n_sim = len(data)
    n_tot = state_ranges[-1][-1] + 1
    df_adjacent, df_err_adjacent = _calculate_df_adjacent(data, df_method)
    df, df_err, overlap_bool = _calculate_weighted_df(df_adjacent, df_err_adjacent, state_ranges, propagated_err=True)

    if err_method == 'bootstrap':
        if seed is not None:
            print(f'Setting the random seed for boostrapping: {seed}')

        # Recalculate err with bootstrapping. (df is still the same and has been calculated above.)
        df_bootstrap = []
        sampled_data_all = [data[i].sample(n=len(data[i]) * n_bootstrap, replace=True, random_state=seed) for i in range(n_sim)]  # noqa: E501
        for b in range(n_bootstrap):
            sampled_data = [sampled_data_all[i].iloc[b * len(data[i]):(b + 1) * len(data[i])] for i in range(n_sim)]
            df_adjacent, df_err_adjacent = _calculate_df_adjacent(sampled_data, df_method)
            df_sampled, _, overlap_bool = _calculate_weighted_df(df_adjacent, df_err_adjacent, state_ranges, propagated_err=False)  # noqa: E501
            df_bootstrap.append(df_sampled)
        error_bootstrap = np.std(df_bootstrap, axis=0, ddof=1)

        # Replace the value in df_err with value in error_bootstrap if df_err corresponds to
        # the df between overlapping states
        for i in range(n_tot - 1):
            if overlap_bool[i] is True:
                print(f'Replaced the propagated error with the bootstrapped error for states {i} and {i + 1}: {df_err[i]:.5f} -> {error_bootstrap[i]:.5f}.')  # noqa: E501
                df_err[i] = error_bootstrap[i]

    df.insert(0, 0)
    df_err.insert(0, 0)
    f = [sum(df[:(i + 1)]) for i in range(len(df))]
    f_err = [np.sqrt(sum([x**2 for x in df_err[:(i+1)]])) for i in range(len(df_err))]

    return f, f_err


def plot_free_energy(f, f_err, fig_name):
    """
    Plot the free energy profile with error bars.

    Parameters
    ----------
    f : list
        The full-range free energy profile.
    f_err : list
        The uncertainty corresponding to the values in :code:`f`.
    fig_name : str
        The file name of the png file to be saved (with the extension).
    """
    plt.figure()
    plt.plot(range(len(f)), f, 'o-', c='#1f77b4')
    plt.errorbar(range(len(f)), f, yerr=f_err, fmt='o', capsize=2, c='#1f77b4')
    plt.xlabel('State')
    plt.ylabel('Free energy (kT)')
    plt.grid()
    plt.savefig(f'{fig_name}', dpi=600)


def average_weights(g_vecs, frac):
    """
    Average the differences between the weights of the coupled and uncoupled states.
    This can be an estimate of the free energy difference between two end states.

    Parameters
    ----------
    g_vecs : np.array
        An array of alchemical weights of the whole range of states as a function of
        simulation time, which is typically generated by :obj:`.combine_weights`.
    frac : float
        The fraction of g_vecs to average over. frac=0.2 means average the last 20% of
        the weight vectors.

    Returns
    -------
    dg_avg : float
        The averaged difference in the weights between the coupled and uncoupled states.
    dg_avg_err : float
        The error of :code:`dg_avg`.
    """
    N = len(g_vecs)
    dg = []
    for i in range(N):
        dg.append(g_vecs[i][-1] - g_vecs[i][0])
    n = int(np.floor(N * frac))
    dg_avg = np.mean(dg[-n:])
    dg_avg_err = np.std(dg_avg[-n:], ddof=1)

    return dg_avg, dg_avg_err



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
Unit tests for the module analyze_traj.py.
"""
import os
from ensemble_md.analysis import analyze_traj

current_path = os.path.dirname(os.path.abspath(__file__))
input_path = os.path.join(current_path, "data")


def test_extract_state_traj():
    traj = analyze_traj.extract_state_traj(os.path.join(input_path, 'dhdl/dhdl_0.xvg'))
    state_list = [
        0, 0, 3, 1, 4, 4, 5, 4, 5, 5,
        4, 4, 5, 4, 2, 4, 5, 2, 1, 2,
        3, 1, 2, 4, 1, 0, 2, 4, 3, 2,
        1, 3, 3, 4, 2, 3, 1, 1, 0, 1,
        2, 3, 1, 0, 1, 4, 3, 1, 3, 2, 5]
    assert traj == state_list


def test_stitch_trajs():
    pass


def test_traj2transmtx():
    pass


def test_plot_rep_trajs():
    pass


def test_plot_state_trajs():
    pass


def test_plot_state_hist():
    pass


def plot_transit_time():
    pass



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
import os
import sys
import glob
import time
import copy
import shutil
import argparse
import numpy as np
from mpi4py import MPI
from datetime import datetime

from ensemble_md.utils import utils
from ensemble_md.ensemble_EXE import EnsembleEXE
from ensemble_md.utils.exceptions import ParameterError


def initialize(args):
    parser = argparse.ArgumentParser(
        description='This code runs an ensemble of expanded ensemble given necessary inputs.')
    parser.add_argument('-y',
                        '--yaml',
                        type=str,
                        default='params.yaml',
                        help='The input YAML file that contains EEXE parameters. (Default: params.yaml)')
    parser.add_argument('-c',
                        '--ckpt',
                        type=str,
                        default='rep_trajs.npy',
                        help='The NPY file containing the replica-space trajectories. This file is a \
                            necessary checkpoint file for extending the simulaiton. (Default: rep_trajs.npy)')
    parser.add_argument('-g',
                        '--g_vecs',
                        type=str,
                        default='g_vecs.npy',
                        help='The NPY file containing the timeseries of the whole-range alchemical weights. \
                            This file is a necessary input if ones wants to update the file when extending \
                            the simulation. (Default: g_vecs.npy)')
    parser.add_argument('-o',
                        '--output',
                        type=str,
                        default='run_EEXE_log.txt',
                        help='The output file for logging how replicas interact with each other. \
                            (Default: run_EEXE_log.txt)')
    parser.add_argument('-m',
                        '--maxwarn',
                        type=int,
                        default=0,
                        help='The maximum number of warnings in parameter specification to be ignored.')
    args_parse = parser.parse_args(args)

    return args_parse


def main():
    t1 = time.time()
    args = initialize(sys.argv[1:])
    sys.stdout = utils.Logger(logfile=args.output)
    sys.stderr = utils.Logger(logfile=args.output)

    # Step 1: Set up MPI rank and instantiate EnsembleEXE to set up EEXE parameters
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()  # Note that this is a GLOBAL variable

    if rank == 0:
        print(f'Current time: {datetime.now().strftime("%d/%m/%Y %H:%M:%S")}')
        print(f'Command line: {" ".join(sys.argv)}\n')

    EEXE = EnsembleEXE(args.yaml)

    if rank == 0:
        # Print out simulation parameters
        EEXE.print_params()

        # Print out warnings and fail if needed
        for i in EEXE.warnings:
            print()
            print(f'{i}')
            print()

        if len(EEXE.warnings) > args.maxwarn:
            raise ParameterError(
                f"The execution failed due to warning(s) about parameter spcificaiton. Check the warnings, or consider setting maxwarn in the input YAML file if you find them harmless.")  # noqa: E501, F541

    # Step 2: If there is no checkpoint file found/provided, perform the 1st iteration (index 0)
    if os.path.isfile(args.ckpt) is False:
        # 2-1. Set up input files for all simulations with 1 rank
        if rank == 0:
            for i in range(EEXE.n_sim):
                os.mkdir(f'sim_{i}')
                os.mkdir(f'sim_{i}/iteration_0')
                MDP = EEXE.initialize_MDP(i)
                MDP.write(f"sim_{i}/iteration_0/{EEXE.mdp.split('/')[-1]}", skipempty=True)
                shutil.copy(f'{EEXE.gro}', f"sim_{i}/iteration_0/{EEXE.gro.split('/')[-1]}")
                shutil.copy(f'{EEXE.top}', f"sim_{i}/iteration_0/{EEXE.top.split('/')[-1]}")

        # 2-2. Run the first ensemble of simulations
        md = EEXE.run_EEXE(0)

        # 2-3. Restructure the directory (move the files from mdrun_0_i0_* to sim_*/iteration_0)
        if rank == 0:
            work_dir = md.output.directory.result()
            for i in range(EEXE.n_sim):
                if EEXE.verbose is True:
                    print(f'  Moving files from {work_dir[i].split("/")[-1]}/ to sim_{i}/iteration_0/ ...')
                    print(f'  Removing the empty folder {work_dir[i].split("/")[-1]} ...')
                for f in glob.glob(f'{work_dir[i]}/*'):
                    shutil.move(f, f'sim_{i}/iteration_0/')
                os.rmdir(work_dir[i])
        start_idx = 1
    else:
        if rank == 0:
            # If there is a checkpoint file, we see the execution as an extension of an EEXE simulation
            ckpt_data = np.load(args.ckpt)
            start_idx = len(ckpt_data[0])
            print(f'\nGetting prepared to extend the EEXE simulation from iteration {start_idx} ...')

            print('Deleting corrupted data ...')
            corrupted = glob.glob('gmxapi.commandline.cli*')  # corrupted iteration
            corrupted.extend(glob.glob('mdrun*'))
            for i in corrupted:
                shutil.rmtree(i)
            if len(corrupted) == 0:
                corrupt_bool = False

            for i in range(EEXE.n_sim):
                n_finished = len(next(os.walk(f'sim_{i}'))[1])  # number of finished iterations (the last might be initialized but corrupted though)  # noqa: E501
                if n_finished == EEXE.n_iter and corrupt_bool is False:
                    print('Extension aborted: The expected number of iterations have been completed!')
                    sys.exit()
                else:
                    print('Deleting data generated after the checkpoint ...')
                    for j in range(start_idx, n_finished):
                        print(f'  Deleting the folder sim_{i}/iteration_{j}')
                        shutil.rmtree(f'sim_{i}/iteration_{j}')

            # Read g_vecs.npy and rep_trajs.npy so that new data can be appended, if any.
            EEXE.rep_trajs = [list(i) for i in ckpt_data]
            if os.path.isfile(args.g_vecs) is True:
                EEXE.g_vecs = [list(i) for i in np.load(args.g_vecs)]
        else:
            start_idx = None

        start_idx = comm.bcast(start_idx, root=0)  # so that all the ranks are aware of start_idx

    for i in range(start_idx, EEXE.n_iter):
        if rank == 0:
            # Step 3: Swap the coordinates
            # 3-1. For all the replica simulations,
            #   (1) Find the last sampled state and the corresponding lambda values from the DHDL files.
            #   (2) Find the final Wang-Landau incrementors and weights from the LOG files.
            dhdl_files = [f'sim_{j}/iteration_{i - 1}/dhdl.xvg' for j in range(EEXE.n_sim)]
            log_files = [f'sim_{j}/iteration_{i - 1}/md.log' for j in range(EEXE.n_sim)]
            states = EEXE.extract_final_dhdl_info(dhdl_files)
            wl_delta, weights, counts = EEXE.extract_final_log_info(log_files)
            print()

            # 3-2. Identify swappable pairs, propose swap(s), calculate P_acc, and accept/reject swap(s)
            # Note after `get_swapping_pattern`, `states` won't be necessarily since it is updated by
            # `get_swapping_pattern`. (Even if the function does not explicitly returns `states`, `states`
            # can still be different after the use of the function.) Therefore, here we create copy of states
            # before the use of `get_swapping_pattern`, so we can use the copy in `update_MDP`.
            states_copy = copy.deepcopy(states)
            swap_pattern = EEXE.get_swapping_pattern(dhdl_files, states, weights)

            # 3-3. Perform histogram correction for the weights as needed
            weights = EEXE.histogram_correction(weights, counts)

            # 3-4. Combine the weights. Note that this is just for initializing the next iteration and is indepdent of swapping itself.  # noqa: E501
            weights, g_vec = EEXE.combine_weights(weights, method=EEXE.w_scheme)
            EEXE.g_vecs.append(g_vec)

            # 3-5. Modify the MDP files and swap out the GRO files (if needed)
            # Here we keep the lambda range set in mdp the same across different iterations in the same folder but swap out the gro file  # noqa: E501
            # Note we use states_copy instead of states in update_MDP.
            for j in list(range(EEXE.n_sim)):
                os.mkdir(f'sim_{j}/iteration_{i}')
                MDP = EEXE.update_MDP(f"sim_{j}/iteration_{i - 1}/{EEXE.mdp.split('/')[-1]}", j, i, states_copy, wl_delta, weights)   # modify with a new template  # noqa: E501
                MDP.write(f"sim_{j}/iteration_{i}/{EEXE.mdp.split('/')[-1]}", skipempty=True)
                shutil.copy(f'{EEXE.top}', f"sim_{j}/iteration_{i}/{EEXE.top.split('/')[-1]}")

                # Now we swap out the GRO files as needed
                shutil.copy(f'sim_{swap_pattern[j]}/iteration_{i-1}/confout.gro', f"sim_{j}/iteration_{i}/{EEXE.gro.split('/')[-1]}")  # noqa: E501

        # Step 4: Perform another iteration
        # 4-1. Run another ensemble of simulations
        md = EEXE.run_EEXE(i)

        if rank == 0:
            # 4-2. Restructure the directory (move the files from mdrun_{i}_i0_* to sim_*/iteration_{i})
            work_dir = md.output.directory.result()
            for j in range(EEXE.n_sim):
                if EEXE.verbose is True:
                    print(f'  Moving files from {work_dir[j].split("/")[-1]}/ to sim_{j}/iteration_{i}/ ...')
                    print(f'  Removing the empty folder {work_dir[j].split("/")[-1]} ...')
                for f in glob.glob(f'{work_dir[j]}/*'):
                    shutil.move(f, f'sim_{j}/iteration_{i}/')
                os.rmdir(work_dir[j])

            # 4-3. Checkpoint as needed
            if (i + 1) % EEXE.n_ckpt == 0:
                print('\n----- Saving .npy files to checkpoint the simulation ---')
                if EEXE.g_vecs[0] is not None:
                    np.save('g_vecs.npy', EEXE.g_vecs)
                np.save('rep_trajs.npy', EEXE.rep_trajs)

    # Save the npy files at the end of the simulation anyway.
    if rank == 0:
        if EEXE.g_vecs[0] is not None:
            np.save('g_vecs.npy', EEXE.g_vecs)
        np.save('rep_trajs.npy', EEXE.rep_trajs)

    # Step 5: Write a summary for the simulation ensemble
    if rank == 0:
        print('\nSummary of the simulation ensemble')
        print('==================================')
        print('Simulation status:')
        for i in range(EEXE.n_sim):
            if EEXE.fixed_weights is True:
                print(f'- Rep {i}: The weights were fixed throughout the simulation.')
            elif EEXE.equil[i] == -1:
                print(f'  - Rep {i}: The weights have not been equilibrated.')
            else:
                idx = int(np.floor(EEXE.equil[i] / (EEXE.dt * EEXE.nst_sim)))
                if EEXE.equil[i] > 1000:
                    units = 'ns'
                    EEXE.equil[i] /= 1000
                else:
                    units = 'ps'
                print(f'  - Rep {i}: The weights have been equilibrated at {EEXE.equil[i]} {units} (iteration {idx}).')

        print(f'\n{EEXE.n_rejected} out of {EEXE.n_swap_attempts}, or {EEXE.n_rejected / EEXE.n_swap_attempts * 100:.1f}% of attempted exchanges were rejected.')  # noqa: E501

        print(f'\nTime elapsed: {utils.format_time(time.time() - t1)}')



####################################################################
#                                                                  #
#    ensemble_md,                                                  #
#    a python package for running GROMACS simulation ensembles     #
#                                                                  #
#    Written by Wei-Tse Hsu <wehs7661@colorado.edu>                #
#    Copyright (c) 2022 University of Colorado Boulder             #
#                                                                  #
####################################################################
"""
The :obj:`.msm_analysis` module provides analysis methods based on Markov state models.
"""
import pyemma
import ruptures as rpt
import matplotlib.pyplot as plt

from ensemble_md.utils import utils


def plot_acf(models, n_tot, fig_name):
    """
    Plots the state index autocorrelation times for all configurations in a single plot

    Parameters
    ----------
    models : list
        A list of MSM models (built by PyEMMA) that have the :code:`correlation` method.
    n_tot : int
        The total number of states (whole range).
    fig_name : str
        The file name of the png file to be saved (with the extension).
    """
    plt.figure()
    for i in range(len(models)):
        if models[i] is not None:
            times, acf = models[i].correlation(models[i].active_set)
            plt.plot(times, acf, label=f'Configuration {i}')
    plt.xlabel('Time')   # Need to figure out what exactly this is ...
    plt.ylabel('Autocorrelation function (ACF)')
    plt.grid()
    plt.legend()
    plt.tight_layout()
    plt.savefig(fig_name, dpi=600)


def plot_its(trajs, lags, fig_name, dt=1, units='step'):
    """
    Plots the implied timescales as a function of lag time for all configurations
    in a subplot.

    Parameters
    ----------
    trajs : list
        A list of state-space trajectories.
    lags : list
        A list of lag times to examine.
    fig_name : str
        The file name of the png file to be saved (with the extension).
    dt : float
        Physical time between frames. The default is 1.
    units : str
        The units of dt. The default is 'ps'.

    Returns
    -------
    ts_list : list
        An list of instances of the ImpliedTimescales class in PyEMMA.
    """
    ts_list = []
    n_rows, n_cols = utils.get_subplot_dimension(len(trajs))
    fig = plt.figure(figsize=(3 * n_cols, 2.5 * n_rows))
    for i in range(len(trajs)):
        # We convert trajs[i] to list to avoid BufferError: memoryview: underlying buffer is not C-contiguous
        ts = pyemma.msm.its(list(trajs[i]), lags=lags, show_progress=False)
        ts_list.append(ts)

        fig.add_subplot(n_rows, n_cols, i + 1)
        pyemma.plots.plot_implied_timescales(ts, dt=dt, units=units)
        plt.xlabel(f'Lag time ({units})')
        plt.ylabel(f'Implied timescale ({units})')
        plt.title(f'Configuration {i}', fontweight='bold')
        plt.grid()

    plt.tight_layout()
    plt.savefig(fig_name, dpi=600)

    return ts_list


def decide_lagtimes(ts_list):
    """
    This function automatically estimates a lagtime for building an MSM for each configuration.
    Specifically, the lag time will be estimated by the change point detection enabled by
    ruptures for each (n-1) timescales (where n is the number of states). A good lag time
    should be long enough such that the timescale is roughly constant but short enough to be
    smaller than all timescales. If no lag time is smaller than all timescales, then a
    warning will be printed and a lag time of 1 will be returned in chosen_lags.

    Parameters
    ----------
    ts_list : list
        An list of instances of the ImpliedTimescales class in PyEMMA.

    Returns
    -------
    chosen_lags: list
        A list of lag time automatically determined for each configuration.
    """
    # Workflow: first find the timescales larger than the corressponding lag times,
    # then perform change change detection.
    chosen_lags = []
    print('     Suggested lag times (in trajectory frames) for each timescale curve of each configuration:')
    for i in range(len(ts_list)):   # for each configuration
        lag_list = []   # a list of lags chosen based on each timescale cure
        ts = ts_list[i]
        for j in range(len(ts.timescales[0])):  # compare each timescale curve with lag times
            ts_arr = ts.timescales[:, j]  # the j-th timescale curve, length: number of lagtimes
            ts_sub = ts_arr[ts_arr > ts.lagtimes]  # timescales that are larger than the corresponding lag times
            if len(ts_sub) <= 10:   # i.e. most timescales < lag time --> no appropirate lag time anyway, use 1
                lag_list.append(1)
            else:
                algo = rpt.Window(width=10, model='l2').fit(ts_sub)
                change_loc = algo.predict(n_bkps=1)  # this returns indices
                lag_list.append(ts.lagtimes[change_loc[0]])  # not sure if the first change point makes sense. Need to check.  # noqa: E501
        print(f'     - Configuration {i}: {lag_list}')  # noqa: E501

        # There might be cases like [6, 1, 1, 1] but using 6 is probably equally bad as 1.
        # If all are larger than one, using the max at least ensure all timescales are roughly constant.
        chosen_lags.append(max(lag_list))  # units: time frame

    return chosen_lags



"""
ensemble_md
A repository for developing ensemble simulation methods
"""
import sys
from setuptools import setup, find_packages
import versioneer

short_description = "A package for setting up, performing, and analyzing molecular dynamics ensembles using GROMACS".split("\n")[0]

# from https://github.com/pytest-dev/pytest-runner#conditional-requirement
needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []

try:
    with open("README.md", "r") as handle:
        long_description = handle.read()
except FileNotFoundError:
    long_description = "\n".join(short_description[2:])


setup(
    # Self-descriptive entries which should always be present
    name='ensemble_md',
    author='Wei-Tse Hsu',
    author_email='wehs7661@colorado.edu',
    description=short_description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license='MIT',
    project_urls={
        "Documentation": "https://ensemble-md.readthedocs.io/",
        "Source Code": "https://github.com/wehs7661/ensemble_md",
    },
    keywords="molecular mechanics, free energy calculations, advanced sampling",

    # Describes the project using a list of classifiers (https://pypi.org/classifiers/)
    # This makes the project more searchable.
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows ",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],

    # Which Python importable modules should be included when your package is installed
    # Handled automatically by setuptools. Use 'exclude' to prevent some specific
    # subpackage(s) from being added, if needed
    packages=find_packages(),

    # Optional include package data to ship with your package
    # Customize MANIFEST.in if the general case does not suit your needs
    # Comment out this line to prevent the files from being packaged with your software
    include_package_data=True,

    # Allows `setup.py test` to work correctly with pytest
    setup_requires=[] + pytest_runner,

    # Add entry points
    entry_points={
        'console_scripts':[
            'run_EEXE = ensemble_md.cli.run_EEXE:main',
            'analyze_EEXE = ensemble_md.cli.analyze_EEXE:main',
            'explore_EEXE = ensemble_md.cli.explore_EEXE:main',
        ],
    },

    # Additional entries you may want simply uncomment the lines you want and fill in the data
    # url='https://github.com/wehs7661/ensemble_md',  # Website

    # Required packages, pulls from pip if needed; do not use for Conda deployment
    install_requires=[
        'numpy',
        'natsort',
        'argparse',
        'pymbar>=4.0.1',
        'alchemlyb>=2.0.0',
        'pyyaml',
        'seaborn',
        'matplotlib',
        'pyemma',
    ],

    # The following extra_require directive provides optional dependencies by, in our case, pip install ensemble[gmxapi].
    # ensemble_md requires GROMACS and gmxapi to be installed before use. 
    # If a working version of GROMACS is available, the user can choose to install ensemble_md along with gmxapi 
    # using the following command: `pip install ensemble_md[gmxapi]`. Otherwise, follow installation 
    # instructions of GROMACS and gmxapi to install each package separately. 
    extras_require={
        'gmxapi': [
            'pybind11>=2.6',
            'setuptools>=42.0',
            'mpi4py',
            'gmxapi>=0.4.0rc2'
        ],
    },

    platforms=['Linux',
               'Mac OS-X',
               'Unix',
               'Windows'],            # Valid platforms your code works on, adjust to your flavor
    
    python_requires=">=3.8",          # Python version restrictions

    # Manual control if final package is compressible or not, set False to prevent the .egg from being made
    # zip_safe=False,

)
